/* This file was generated by syncqt. */
#ifndef QT_QTQMLMETA_VERSION_H
#define QT_QTQMLMETA_VERSION_H

#define QTQMLMETA_VERSION_STR "6.8.3"

#define QTQMLMETA_VERSION 0x060803

#endif // QT_QTQMLMETA_VERSION_H
