// Copyright (C) 2019 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#ifndef QTQMLMODELSGLOBAL_H
#define QTQMLMODELSGLOBAL_H

#include <QtQml/qtqmlglobal.h>
#include <QtQmlModels/qtqmlmodels-config.h>
#include <QtQmlModels/qtqmlmodelsexports.h>

#endif // QTQMLMODELSGLOBAL_H
