/* This file was generated by syncqt. */
#ifndef QT_QTQMLCOMPILER_VERSION_H
#define QT_QTQMLCOMPILER_VERSION_H

#define QTQMLCOMPILER_VERSION_STR "6.8.3"

#define QTQMLCOMPILER_VERSION 0x060803

#endif // QT_QTQMLCOMPILER_VERSION_H
